#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <xc.h>
#include "uart.h"
#include "control.h"

int upperGen = 4.8;
int lowerGen = 0;
int upperLoad = 4;
int lowerLoad = 0;
int lowerBatt = 10;
int upperBatt = 13;
//voltage divider resistor values
double R = 250000, r = 2500, shuntR = 10;

/*UART read in and calculations*/
double convertAnalogReading(double);

//read in V for batt and related calculation variables
double vBatt1, vBatt2, batteryVoltage, batteryCurrent, powerAtBattery, prevBattCurrent = 0.1; 

//read in V for LOAD and related calculation variables
double vLoad1_1, vLoad1_2, load1Voltage, load1Current, powerAtLoad1, prevLoad1Current = 0.01; 
double vLoad2_1, vLoad2_2, load2Voltage, load2Current, powerAtLoad2, prevLoad2Current = 0.01; 

//read in V for GENERATION and related calculation variables
double vGen1_1, vGen1_2, vGen2_1, vGen2_2;
double gen1Voltage, gen2Voltage, gen1Current, gen2Current;
double powerAtGen1, powerAtGen2, prevGen1Current = 0.1, prevGen2Current = 0.1; 
double prevGen1Voltage = 0, prevGen2Voltage = 0;
int dischargeFlag = 2;
int Gen1Flag = 0;
int Gen2Flag = 0;
int nightFlag = 0;


double convertAnalogReading(double analogIn){
    double convertedValue = (analogIn/4095)*3.3;
    return convertedValue;
}


void initSSR(void){
    
    // Battery outflow switch (battery supplies power to loads)
    TRISBbits.TRISB5 = 0;
    LATBbits.LATB5 = 0;
    // Battery inflow switch (charge battery)
    TRISBbits.TRISB6 = 0;
    LATBbits.LATB6 = 0;
    batteryOutflowSwitch = open; //nothing connected upon initial startup
    batteryInflowSwitch = open;
}

void dischargeBatt(void) {
    // discharging
    batteryInflowSwitch = open;
    batteryOutflowSwitch = close;
    dischargeFlag = 1;
}

void chargeBatt(void) {
    // charging
    batteryOutflowSwitch = open;
    batteryInflowSwitch = close;
    dischargeFlag = 0;
}

void openSSR(void){
    batteryOutflowSwitch = open; //nothing connected upon initial startup
    batteryInflowSwitch = open; 
    dischargeFlag = 2;
}

void initTimer2(void){
    asm("ei");  //enable global interrupts
    enableMultivectorInterrupts; // multi-vectored interrupts on
    setTimer2InterruptPriority = 3;  //timer2 interrupt priority set to 3
    clearTimer2InterruptFlag; // Clear interrupt flag to start
    enableTimer2Interrupt;
    timer2On;
    timer2Prescale = 0b110; //64 prescale
    PR2 = 39999;
    timer2Interrupt = 0;
    TMR2 = 0;
}

void __attribute__((vector(_TIMER_2_VECTOR), interrupt(ipl3soft), nomips16)) _T2Interrupt(void) {
    int j;
    char Va, Vb, Vc, Vd, Ve, Vf, Vg, Vh, Vi, Vj, Vbd;
    //////***UART RECEIVE***//////
        // read in data transmitted data from Power Monitoring UART
    char x = getu1(); // check current character
    while(x != ','){ // wait until current char is ',' for load 1 start char
        x = getu1();
    }
        Vc = getu1();
        Vd = getu1(); 
        // send data out to display system 
        putu2(','); // send start character
        printf("\r\n"); // send break so display can read
        putu2(Vc);
        printf("\r\n"); // send break so display can read
        putu2(Vd);
        printf("\r\n"); // send break so display can read

    while(x != '-'){ // wait until current char is '-' for generation 1 start char
        x = getu1();
    }
        Vg = getu1();
        Vh = getu1(); 
        // send data out to display system 
        putu2('-'); // send start character
        putu2(Vg);
        printf("\r\n"); // send break so display can read
        putu2(Vh);
        printf("\r\n"); // send break so display can read
        
    while(x != '0'){ // wait until current char is '0' for blocking diode reading
        x = getu1();
    }
        Vbd = getu1();
        
    while(x != '.'){ // wait until current char is '.' for load 2 start char
        x = getu2();
    }
        Ve = getu2();
        Vf = getu2(); 
        // send data out to display system 
        putu2('.'); // send start character
        putu2(Ve);
        printf("\r\n"); // send break so display can read
        putu2(Vf);
        printf("\r\n"); // send break so display can read
        
    while(x != '/'){ // wait until current char is '/' for generation 2 start char
        x = getu2();
    }
        Vi = getu2();
        Vj = getu2(); 
        // send data out to display system 
        putu2('/'); // send start character
        putu2(Vi);
        printf("\r\n"); // send break so display can read
        putu2(Vj);
        printf("\r\n"); // send break so display can read


  

    
    //* Read Battery Voltages *//
    AD1CON1bits.SAMP = 0; // set this bit to zero in our sample loop when we want sampling to stop and conversion to begin
    Nop_delay(); //give the ADC time to convert the value it reads in
    AD1CON1bits.SAMP = 1; // start sampling again
     Va = ADC1BUF12; // get first battery voltage
     Vb = ADC1BUF13; // get second battery voltage
     putu2('+'); // send start character
     putu2(Va);
     printf("\r\n"); // send break so display can read
     putu2(Vb);
     printf("\r\n"); // send break so display can read
        
    //*Voltage, Currents, and Power Calculations*//
        //BATTERY (Va and Vb are the mV range voltages that we read in via UART from batt)
    double VB1 = (double)Va; 
    double VB2 = (double)Vb;
    if ((double)Va < 0 ){ // scale up by 256 to get out of negative range
        VB1 = VB1 + 256;
    }
    if ((double)Vb < 0 ){ // scale up by 256 to get out of negative range
        VB2 = VB2 + 256;
    }   
        vBatt1 = ((VB1/4095)*3.3)*((R+r)/r); //voltage divider: R = 250k, r = 2.5k, global doubles
        vBatt2 = ((VB2/4095)*3.3)*((R+r)/r);
        batteryVoltage = vBatt1;
        if (vBatt1-vBatt2 < 0) {
            batteryCurrent = (vBatt2-vBatt1)/shuntR;
        }
        else if (vBatt1-vBatt2 > 0) {
            batteryCurrent = (vBatt1-vBatt2)/shuntR;
        }
        else {
            batteryCurrent = prevBattCurrent;
        }
        prevBattCurrent = batteryCurrent;
        powerAtBattery = batteryVoltage*batteryCurrent;
        
    /*******************************************************/
        //LOAD 1 - moderate & low-demand load (Vc and Vd are the mV range voltages that we read in via UART from across load(s)1 )
    double V1_1 =  (double)Vc; 
    double V1_2 =  (double)Vd;
    if ((double)Vc < 0 ){ // scale up by 256 to get out of negative range
        V1_1 = V1_1 + 256;
    }
    if ((double)Vd < 0 ){ // scale up by 256 to get out of negative range
        V1_2 = V1_2 + 256;
    }
        vLoad1_1 = ((V1_1/4095)*3.3)*((R+r)/r); //voltage divider: R = 250k, r = 2.5k, global doubles
        vLoad1_2 = ((V1_2/4095)*3.3)*((R+r)/r);
        load1Voltage = vLoad1_1;
        if (vLoad1_1-vLoad1_2 < 0) {
            load1Current = (vLoad1_2-vLoad1_1)/shuntR;
        }
        else if (vLoad1_1-vLoad1_2 > 0) {
            load1Current = (vLoad1_1-vLoad1_2)/shuntR;
        }
        else {
            load1Current = prevLoad1Current;
        }
        prevLoad1Current = load1Current;
        powerAtLoad1 = load1Voltage*load1Current;
        /*******************************************************/
        
    //GENERATION 1 (Ve and Vf are the mV range voltages that we read in via UART from batt)
            //GEN 1
        double VG1_1 =  (double)Vg; 
        double VG1_2 =  (double)Vh;
        if ((double)Vg < 0 ){ // scale up by 256 to get out of negative range
          VG1_1 = VG1_1 + 256;
        }
        if ((double)Vh < 0 ){ // scale up by 256 to get out of negative range
          VG1_2 = VG1_2 + 256;
        }
        vGen1_1 = ((VG1_1/4095)*3.3)*((R+r)/r); //voltage divider: R = 250k, r = 2.5k, global doubles
        vGen1_2 = ((VG1_2/4095)*3.3)*((R+r)/r);
        
        double vGen1Diff = vGen1_1-prevGen1Voltage;
        if (prevGen1Voltage == 0) {
            prevGen1Voltage = vGen1_1;
            gen1Voltage = vGen1_1;
            
        }
        else if (vGen1Diff<0.2 && vGen1Diff>-0.2) {
            if (Gen1Flag == 1) {
                gen1Voltage = vGen1_1;
                Gen1Flag = 0;
            }
        }
        else if (vGen1Diff>6 && vGen1Diff<-6) {
            Gen1Flag = 1;
            prevGen1Voltage = vGen1_1;
        }
        else {
            gen1Voltage = vGen1_1;
        }
                
        
        if (vGen1_1-vGen1_2 < 0) {
            gen1Current = (vGen1_2-vGen1_1)/shuntR;
        }
        else if (vGen1_1-vGen1_2 > 0) {
            gen1Current = (vGen1_1-vGen1_2)/shuntR;
        }
        else {
            gen1Current = prevGen1Current;
        }
        prevGen1Current = gen1Current;
        powerAtGen1 = gen1Voltage*gen1Current;    
        
    /*******************************************************/
    //LOAD 2 - 1.5W DC Motor (Ve and Vf are the mV range voltages that we read in via UART from across load 2 )
    double V2_1 =  (double)Ve; 
    double V2_2 =  (double)Vf;
    if ((double)Ve < 0 ){ // scale up by 256 to get out of negative range
        V2_1 = V2_1 + 256;
    }
    if ((double)Vf < 0 ){ // scale up by 256 to get out of negative range
        V2_2 = V2_2 + 256;
    }
        vLoad2_1 = ((V2_1/4095)*3.3)*((R+r)/r); //voltage divider: R = 250k, r = 2.5k, global doubles
        vLoad2_2 = ((V2_2/4095)*3.3)*((R+r)/r);
        load2Voltage = vLoad2_1;
        if (vLoad2_1-vLoad2_2 < 0) {
            load2Current = (vLoad2_2-vLoad2_1)/shuntR;
        }
        else if (vLoad2_1-vLoad2_2 > 0) {
            load2Current = (vLoad2_1-vLoad2_2)/shuntR;
        }
        else {
            load2Current = prevLoad2Current;
        }
        prevLoad2Current = load2Current;
        powerAtLoad2 = load2Voltage*load2Current;
        /*******************************************************/
        
   //GENERATION 2 (Vi and Vj are the mV range voltages that we read in via UART from batt)
            //GEN 1
        double VG2_1 =  (double)Vi; 
        double VG2_2 =  (double)Vj;
        if ((double)Vi < 0 ){ // scale up by 256 to get out of negative range
          VG2_1 = VG2_1 + 256;
        }
        if ((double)Vj < 0 ){ // scale up by 256 to get out of negative range
          VG2_2 = VG2_2 + 256;
        }
        vGen2_1 = ((VG2_1/4095)*3.3)*((R+r)/r); //voltage divider: R = 250k, r = 2.5k, global doubles
        vGen2_2 = ((VG2_2/4095)*3.3)*((R+r)/r);
        
        double vGen2Diff = vGen2_1-prevGen2Voltage;
        if (prevGen2Voltage == 0) {
            prevGen2Voltage = vGen2_1;
            gen2Voltage = vGen2_1;
        }
        else if (vGen2Diff<0.2 && vGen2Diff>-0.2) {
            if (Gen2Flag == 1) {
                gen2Voltage = vGen2_1;
                Gen2Flag = 0;
            }
        }
        else if (vGen2Diff>6 && vGen2Diff<-6) {
            Gen2Flag = 1;
            prevGen2Voltage = vGen2_1;
        }
        else {
            gen2Voltage = vGen2_1;
        }        
        
        
        if (vGen2_1-vGen2_2 < 0) {
            gen2Current = (vGen2_2-vGen2_1)/shuntR;
        }
        else if (vGen2_1-vGen2_2 > 0) {
            gen2Current = (vGen2_1-vGen2_2)/shuntR;
        }
        else {
            gen2Current = prevGen2Current;
        }
        prevGen2Current = gen2Current;
        powerAtGen2 = gen2Voltage*gen2Current;        
       
        
        
    // Blocking Diode Reading - For determining if Sun Lamp is on or off
    double blockingV = (double)Vbd;
    if ((double)Vbd < 0 ){ // scale up by 256 to get out of negative range
        blockingV = blockingV + 256;
    }
    if (blockingV < 30){ // if value is less than 80, blocking diode isn't dropping much voltage and sun lamp is off
        nightFlag = 1;
    }
    else {               // if it's not off, then it's on
        nightFlag = 0;
    }
        
        
  /*******************************************************/
       // Power Switching
        if (((gen1Voltage < 1.2) || (gen2Voltage < 1.2)) && nightFlag == 1) // if sun is off, turn the battery on to power the loads
        {
            dischargeBatt();
        }
        else if ((gen1Voltage < 3.0) && (gen2Voltage < 3.0) && (dischargeFlag != 1)) // voltage of system is to low; switch on the battery to compensate
        {
            dischargeBatt();
        }
        else if (((gen1Voltage > 11.5) || (gen2Voltage > 11.5)) && (dischargeFlag == 2) && (nightFlag == 0)){ // charge the battery; this should only be the case if the local storage is switched in
            chargeBatt();
        }
        else if (((gen1Voltage > 10.5) || (gen2Voltage > 10.5)) && (nightFlag == 0) && (dischargeFlag != 0)) { // in this case the sun is on but no loads are on so we don't need the battery
            openSSR();
        } 
        else if (((gen1Voltage < 5) || (gen2Voltage < 5)) && (nightFlag == 0) && (dischargeFlag == 0)){ // if the battery is done charging, open both switches
            openSSR();
        } 
    
    timer2Interrupt = 0; //clear timer flag
}

