/* File:   uart.c
 * Author: Kelsey
 * Created on April 12, 2021, 1:48 PM
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <xc.h>
#include "uart.h"
#include "delay.h"


void putu2(char c) {
    while(buff_full);//check that there is room in transmit fifo first
    U2TXREG = c;
    for_delay();
}

char getu1(void){
    char c;
    while(!U1STAbits.URXDA){
        if(U1STAbits.OERR == 1){
            U1STAbits.OERR = 0;
        }
    };  //check that there is a character to be read
    c = U1RXREG;
    return c;
}

char getu2(void){
    char c;
    while(!data_ready){
        if(U2STAbits.OERR == 1){
            U2STAbits.OERR = 0;
        }
    };  //check that there is a character to be read
    c = U2RXREG;
    return c;
}

// UART 1 to read in power data from load1 and gen1
void initSerial1(unsigned long rate){  
    
  U1BRG = 102;  // set to 102 for rate of 57600
  U1MODE = 0; // turn off UART
  U1MODEbits.BRGH = 1; // enable high speed setting, because of formula^^
  U1MODEbits.UEN = 1; // enable UART
  U1STAbits.URXEN = 1; //enable receiver
  U1STAbits.UTXEN = 0; // disable transmitting
  U1MODEbits.ON = 1; // turn on UART 
}

void initUART1(void){
    //initClock(); initialized in main right before UART

    ANSELBbits.ANSB14 = 0; // make sure RA0 and RA1 are set to digital for UART1
    ANSELBbits.ANSB15 = 0;
    TRISBbits.TRISB14 = 0; // TX set TRISB bit to 0 for UART1 output 
    TRISBbits.TRISB15 = 1;  // RX set TRISB bit to 1 for UART1 input
   
    TRISBbits.TRISB11 = 0; // make sure mux select is set as output
    LATBbits.LATB11 = 0; // read in data from the first mux input (set to 1 for second)
    
    ANSELBbits.ANSB13 = 0; // make sure B13 set to digital 
    TRISBbits.TRISB13 = 0; // mux enable is set as output
    LATBbits.LATB13 = 1; // enable mux
}


// UART 2 to read in power data from load2 and gen2, and to sent data to display
void initSerial2(unsigned long rate){  
    
  U2BRG = 102;  // set to 102 for rate of 57600
  U2MODE = 0; // turn off UART
  U2MODEbits.BRGH = 1; // enable high speed setting, because of formula^^
  U2MODEbits.UEN = 1; // enable UART
  U2STAbits.URXEN = 1; //enable receiver
  U2STAbits.UTXEN = 1; // enable transmitting
  U2MODEbits.ON = 1; // turn on UART 
}

void initUART2(void){
    //initClock(); initialized in main right before UART

    ANSELAbits.ANSA0 = 0; // make sure RA0 and RA1 are set to digital for UART2
    ANSELAbits.ANSA1 = 0;
    TRISAbits.TRISA0 = 1; // set TRISA bit 0 as input and 1 as output for UART2
    TRISAbits.TRISA1 = 0;
    
    SYSKEY = 0x00000000; // force lock
    SYSKEY = 0xAA996655; // unlock sequence
    SYSKEY = 0x556699AA; // lock sequence
    
    RPINR9bits.U2RXR = 0b0001; // set RP1 (RA0) to U2RX
    RPOR0bits.RP2R = 0b0100; // set RP2 (RA1) to U2TX pg 125
    //^Function 4 is U2TX (page 119)
    
    SYSKEY = 0x00000000; // relock
}


void _mon_putc(char c){ // standard i/o print
    putu2(c);
}

void initClock(void){
    SYSKEY = 0x00000000; // force lock
    SYSKEY = 0xAA996655; // unlock sequence
    SYSKEY = 0x556699AA; // lock sequence
    
    // user code to modify register contents for reference clock (RB15)
    REFO1CONbits.OE = 1; //  Reference oscillator module is enabled 
    REFO1CONbits.ON = 1; // turn it on
    REFO1CONbits.RODIV = 12; 
    REFO1TRIMbits.ROTRIM = 0;
    REFO1CONbits.ROSEL = 0b0000; // set ref clock to SYS CLK (part 6)
 
    // change the PLL multipliers so Fvco = 96 MHZ
    SPLLCONbits.PLLMULT = 0b0000101; // multiply sys clock by 12 (8*12 = 96 = FVCO)
    SPLLCONbits.PLLODIV = 0b010;
    
    // switch the clock source to PLL
    OSCCONbits.NOSC = 0b001; // set new oscillator source to PLL
    OSCCONbits.OSWEN = 1; // start the swtiching sequence
    while(OSCCONbits.OSWEN); // wait for switching to finish
    
    // relock
    SYSKEY = 0x00000000; // relock
}

